@@ -1,7 +1,7 @@
 ''' Read font file and write glyphs to SVG '''
 
 from __future__ import annotations
-from typing import Literal, Union, Optional, Dict
+from typing import Literal, Union, Optional, Sequence, Dict
 import sys
 import os
 import math
@@ -14,37 +14,37 @@ import warnings
 from .config import config
 from .fontread import FontReader
 from . import gpos
+from . import gsub
 from .cmap import Cmap12, Cmap4
-from .glyph import read_glyph, fmt, SimpleGlyph, CompoundGlyph
-from .fonttypes import AdvanceWidth, Layout, Header, Table, FontInfo, FontNames, Symbols
-
-
-fontindex = None
+from .glyph import SimpleGlyph, CompoundGlyph
+from .glyphcff import read_glyph_cff, CFF
+from .glyphglyf import read_glyph_glyf
+from .glyphinspect import ShowGlyphs
+from .fonttypes import AdvanceWidth, Layout, Header, Table, FontInfo, FontNames, Symbols, FontFeatures
+from .svgpath import fmt
 
 
 class Font:
     ''' Class to read/parse a OpenType/TTF and write glyphs to SVG
 
         Args:
-            name: File name of the font
-            style: Font style such as "bold" or "italic", used when searching
-                system paths for a font file
+            name: File name of the font. Defaults to DejaVuSans.
     '''
-    def __init__(self, name: Union[str, Path]=None, style: str='regular'):
+    def __init__(self, name: Union[str, Path] = None):
         self.fname = None
-        if name and Path(name).exists():
-            self.fname = Path(name)
-        elif name:
-            self.fname = findfont(name, style)
-            if self.fname is None:
-                warnings.warn(f'Font {name} not found.')
-
-        if self.fname is None:
+        if name:
+            if Path(name).exists():
+                self.fname = Path(name)
+            else:
+                raise FileNotFoundError(f'Font {name} not found.')
+        else:
             with pkg_resources.path('ziafont.fonts', 'DejaVuSans.ttf') as p:
                 self.fname = p
 
         with open(self.fname, 'rb') as f:
             self.fontfile = FontReader(f.read())
+
+        self.features = FontFeatures()
         self.info = self._loadfont()  # Load in all the font metadata
         self._glyphs: Dict[int, Union[SimpleGlyph, CompoundGlyph]] = {}
         self._glyphids: Dict[str, int] = {}
@@ -62,6 +62,14 @@ class Font:
         if 'GPOS' in self.tables:
             self.gpos = gpos.Gpos(self.tables['GPOS'].offset, self.fontfile)
 
+        self.gsub = None
+        if 'GSUB' in self.tables:
+            self.gsub = gsub.Gsub(self.tables['GSUB'].offset, self.fontfile)
+
+        self.cffdata: Optional[CFF] = None
+        if 'CFF ' in self.tables:
+            self.cffdata = CFF(self)
+
         return info
 
     def _readtables(self) -> None:
@@ -81,10 +89,10 @@ class Font:
                                      offset=self.fontfile.readuint32(),
                                      length=self.fontfile.readuint32())
 
-        if 'glyf' not in self.tables:
-            raise ValueError('Unsupported font (no glyf table).')
+        if 'glyf' not in self.tables and 'CFF ' not in self.tables:
+            raise ValueError('Unsupported font (no glyf or CFF table).')
 
-    def verifychecksum(self):
+    def verifychecksum(self) -> None:
         ''' Verify checksum of all font tables. Raises ValueError if invalid '''
         for table in self.tables.keys():
             if table != 'head':
@@ -137,15 +145,17 @@ class Font:
             self.fontfile.readint16()  # Skip reserved
         metricformat = self.fontfile.readint16()
         numlonghormetrics = self.fontfile.readuint16()
-
         advwidth = AdvanceWidth(advwidthmax, minleftbearing)
 
+        self.fontfile.readuint32(self.tables['maxp'].offset)
+        numglyphs = self.fontfile.readuint16()
+
         layout = Layout(unitsperem, xmin, xmax, ymin, ymax, ascent, descent,
                         advwidth, minleftbearing, minrightbearing)
         header = Header(version, revision, chksumadjust, magic, flags,
                         created, modified, macstyle,
                         lowestrecppem, directionhint, indextolocformat,
-                        glyphdataformat, numlonghormetrics)
+                        glyphdataformat, numlonghormetrics, numglyphs)
         return header, layout
 
     def _readnames(self) -> FontNames:
@@ -206,7 +216,8 @@ class Font:
         self.cmaps: list[Union[Cmap12, Cmap4]] = []
         cmap: Union[Cmap12, Cmap4]
         for ctable in cmaptables:
-            cmapformat = self.fontfile.readuint16(self.tables['cmap'].offset + ctable.offset)
+            ctable_ofst = self.tables['cmap'].offset + ctable.offset
+            cmapformat = self.fontfile.readuint16(ctable_ofst)
             if cmapformat == 4:
                 endcodes = []
                 startcodes = []
@@ -231,7 +242,7 @@ class Font:
 
                 # Length of glyph array comes from total length of cmap table
                 # //2 because len is in bytes, but table glyphidxarray is 16-bit
-                glyphtablelen = (length - (self.fontfile.tell() - self.tables['cmap'].offset)) // 2
+                glyphtablelen = (length - (self.fontfile.tell() - ctable_ofst)) // 2
                 for i in range(glyphtablelen):
                     glyphidxarray.append(self.fontfile.readuint16())
                 cmap = Cmap4(ctable.platform, ctable.platformid,
@@ -264,6 +275,44 @@ class Font:
         ''' Select cmap table by index. Only supported tables are included. '''
         self.cmap = self.cmaps[cmapidx]
 
+    def scripts(self):
+        ''' Get list of scripts in the font '''
+        scripts = set()
+        if self.gpos:
+            scripts = scripts.union(set(self.gpos.scripts.keys()))
+        if self.gsub:
+            scripts = scripts.union(set(self.gsub.scripts.keys()))
+        return list(scripts)
+
+    def languages(self, script='DFLT'):
+        ''' Get list of languages in the script '''
+        langs = set()
+        if self.gpos:
+            s = self.gpos.scripts.get(script, None)
+            if s is not None:
+                langs = langs.union(set(s.languages.keys()))
+            
+        if self.gsub:
+            s = self.gsub.scripts.get(script, None)
+            if s is not None:
+                langs = langs.union(set(s.languages.keys()))
+        return list(langs)
+       
+    def language(self, script, language):
+        ''' Set script/language to use '''
+        if script not in self.scripts():
+            raise ValueError(f'Script {script} not defined in font')
+
+        if language not in self.languages(script):
+            raise ValueError(f'Language {language} not defined in font')
+
+        if self.gpos:
+            self.gpos.language.script = script
+            self.gpos.language.language = language
+        if self.gsub:
+            self.gsub.language.script = script
+            self.gsub.language.language = language
+
     def glyphindex(self, char: str) -> int:
         ''' Get index of character glyph '''
         gid = self._glyphids.get(char)
@@ -271,24 +320,7 @@ class Font:
             gid = self.cmap.glyphid(char)  # type: ignore
             self._glyphids[char] = gid
         return gid
-
-    def _glyphoffset(self, index: int) -> Optional[int]:
-        ''' Get offset (from beginning of file) of glyph,
-            Return None if no glyph (empty/space) at this index.
-        '''
-        if self.info.header.indextolocformat == 1:
-            offset = self.fontfile.readuint32(self.tables['loca'].offset + index * 4)
-            nextofst = self.fontfile.readuint32()
-        else:
-            offset = self.fontfile.readuint16(self.tables['loca'].offset + index * 2) * 2
-            nextofst = self.fontfile.readuint16() * 2
-
-        if offset == nextofst:
-            # Empty glyphs (ie space) have no length.
-            return None
-        else:
-            return offset + self.tables['glyf'].offset
-
+            
     def glyph(self, char: str) -> SimpleGlyph:
         ''' Get the Glyph for the character '''
         index = self.glyphindex(char)        # Glyph Number
@@ -302,18 +334,23 @@ class Font:
         '''
         glyph = self._glyphs.get(glyphid)
         if glyph is None:
-            glyph = read_glyph(glyphid, self)
+            if 'CFF ' in self.tables:
+                glyph = read_glyph_cff(glyphid, self)
+            else:  # 'glyf' table
+                glyph = read_glyph_glyf(glyphid, self)
             self._glyphs[glyphid] = glyph
         return glyph
 
-    def advance(self, glyph1: int, glyph2: int=None, kern: bool=True):
-        ''' Get advance width in font units, including kerning adjustment if glyph2 is defined '''
+    def advance(self, glyph1: int, glyph2: int = None):
+        ''' Get advance width in font units, including kerning adjustment
+            if glyph2 is defined
+        '''
         try:
             adv = self.advwidths[glyph1].width
         except IndexError:
             adv = self.info.layout.advwidthmax.width
 
-        if kern and glyph2 and self.gpos:
+        if self.features.kern and glyph2 and self.gpos:
             # Only getting x-advance for first glyph.
             adv += self.gpos.kern(glyph1, glyph2)[0].get('xadvance', 0)
         return adv
@@ -330,8 +367,7 @@ class Font:
              valign: Literal['base', 'center', 'top']='base',
              color: str=None,
              rotation: float=0,
-             rotation_mode: str='anchor',
-             kern=True):
+             rotation_mode: str='anchor'):
         ''' Create a Text object using this font
 
             Args:
@@ -348,10 +384,12 @@ class Font:
                 kern: Use font kerning adjustment
         '''
         txt = Text(s, self, size=size, linespacing=linespacing, halign=halign, valign=valign,
-                   color=color, kern=kern, rotation=rotation,
-                   rotation_mode=rotation_mode)
+                   color=color, rotation=rotation, rotation_mode=rotation_mode)
         return txt
 
+    def showallglyphs(self, size: float = 36, pxwidth: int = 800):
+        return ShowGlyphs(self)
+
 
 class Text:
     ''' Draw text as SVG paths
@@ -368,16 +406,14 @@ class Text:
             rotation_mode: Either 'default' or 'anchor', to
                 mimic Matplotlib behavoir. See:
                 https://matplotlib.org/stable/gallery/text_labels_and_annotations/demo_text_rotation_mode.html
-            kern: Use kerning adjustment
     '''
-    def __init__(self, s: str,  font: Union[str, Font]=None,
-                 size: float=None, linespacing: float=1,
-                 halign: Literal['left', 'center', 'right']='left',
-                 valign: Literal['base', 'center', 'top']='base',
-                 color: str=None,
-                 rotation: float=0,
-                 rotation_mode: str='anchor',
-                 kern: bool=True):
+    def __init__(self, s: Union[str, Sequence[int]],  font: Union[str, Font] = None,
+                 size: float = None, linespacing: float = 1,
+                 halign: Literal['left', 'center', 'right'] = 'left',
+                 valign: Literal['base', 'center', 'top'] = 'base',
+                 color: str = None,
+                 rotation: float = 0,
+                 rotation_mode: str = 'anchor'):
         self.str = s
         self.halign = halign
         self.valign = valign
@@ -386,7 +422,6 @@ class Text:
         self.rotation_mode = rotation_mode
         self.size = size if size else config.fontsize
         self.linespacing = linespacing
-        self.kern = kern
         if font is None or isinstance(font, str):
             self.font = Font(font)
         else:
@@ -415,14 +450,14 @@ class Text:
         ''' Jupyter representer '''
         return self.svg()
 
-    def drawon(self, svg: ET.Element, x: float=0, y: float=0):
+    def drawon(self, svg: ET.Element, x: float = 0, y: float = 0):
         ''' Draw text on the SVG '''
         svg, _ = self._drawon(svg, x, y)
         return svg
 
-    def _drawon(self, svg: ET.Element, x: float=0, y: float=0):
+    def _drawon(self, svg: ET.Element, x: float = 0, y: float = 0):
         ''' Draw text on the SVG '''
-        word, symbols, width, ymin, ymax = self._symbols
+        word, symbols, width, xmin, ymin, ymax = self._symbols
         height = ymax-ymin
         # Adjust vertical alignment
         yofst = {'base': 0,
@@ -461,7 +496,7 @@ class Text:
             circ.attrib['r'] = '3'
             circ.attrib['fill'] = 'red'
             circ.attrib['stroke'] = 'red'
-        bbox = (xy[0], xy[0]+width, xy[1]+ymin, xy[1]+ymax)
+        bbox = (xy[0]+xmin, xy[0]+width, xy[1]+ymin, xy[1]+ymax)
 
         if self.rotation:
             centerx = xy[0]  # Center of rotation
@@ -510,38 +545,69 @@ class Text:
             word.attrib['fill'] = self.color
         return svg, bbox
 
+    def str_to_gids(self):
+        if isinstance(self.str, str):
+            lines = self.str.splitlines()
+            gidlines = [[self.font.glyphindex(c) for c in line] for line in lines]
+        else:
+            gidlines = [self.str]  # Single line
+        return gidlines
+
     def _buildstring(self) -> Symbols:
         ''' Create symbols and svg word in a <g> group tag, for placing in an svg '''
         scale = self.size / self.font.info.layout.unitsperem
         lineheight = self.size * self.linespacing
-        lines = self.str.splitlines()
-        yvals = [i*lineheight for i in range(len(lines))]  # valign == 'base'
+        
+        gidlines = self.str_to_gids()
+        yvals = [i*lineheight for i in range(len(gidlines))]  # valign == 'base'
 
         # Generate symbols and calculate x positions using left alignment
         symbols: list[ET.Element] = []  # <symbol> elements
         linewidths: list[float] = []
-        allglyphs = []  # (glyph, x) where x is left aligned
-        for lineidx, line in enumerate(lines):
+        xadvance = 0.
+        allglyphs = []  # (glyph, x, y) where x is left aligned, y is delta from baseline
+        xmin = 0
+        for lineidx, glyphids in enumerate(gidlines):
             lineglyphs = []
-            glyphs = [self.font.glyph(c) for c in line]
+            if self.font.gsub:
+                glyphids = self.font.gsub.sub(glyphids, self.font.features)
+
+            glyphs = [self.font.glyph_fromid(gid) for gid in glyphids]
             x = 0
+            xmin = min(xmin, glyphs[0].bbox.xmin*scale)
             for gidx, glyph in enumerate(glyphs):
                 if glyph.id not in [s.attrib['id'] for s in symbols]:
                     symbols.append(glyph.svgsymbol())
-                lineglyphs.append((glyph, x))
+                                
+                prevglyph = glyphs[gidx-1] if gidx-1 >= 0 else None
                 nextglyph = glyphs[gidx+1] if gidx+1 < len(glyphs) else None
-                xadvance = glyph.advance(nextglyph, kern=self.kern)
-                x += (xadvance - min(0, glyph.path.bbox.xmin)) * scale
+                if self.font.gpos and prevglyph:
+                    # Attach marks
+                    placexy = self.font.gpos.placemark(prevglyph.index, glyph.index)
+                    if placexy:
+                        if not placexy.mkmk:  # Relative to base glyph
+                            dx = dy = 0
+                        dx += (placexy.dx - xadvance) * scale
+                        dy += -placexy.dy * scale
+                    else:
+                        dx = dy = 0
+                else:
+                    dx = dy = 0  # Don't reset, may need these for mark-to-mark p[
+                        
+                lineglyphs.append((glyph, x+dx, dy))                
+                xadvance = glyph.advance(nextglyph)
+                x += xadvance * scale
 
-            if glyph.path.bbox.xmax > xadvance:
-                # Make a bit wider to grab right edge that extends beyond advance width
-                x += (glyph.path.bbox.xmax - xadvance) * scale
+            if glyph.bbox.xmax > xadvance:
+                # Make linewidth a bit wider to grab right edge
+                # that extends beyond advance width
+                x += (glyph.bbox.xmax - xadvance) * scale
+                x += (-glyphs[0].bbox.xmin) * scale
             linewidths.append(x)
             allglyphs.append(lineglyphs)
 
         # Place the glyphs based on halign
         word = ET.Element('g')
-        word.attrib['word'] = self.str  # Just an identifier for debugging
         totwidth = max(linewidths)
         for lineidx, (lineglyphs, linewidth) in enumerate(zip(allglyphs, linewidths)):
             if self.halign == 'center':
@@ -550,15 +616,17 @@ class Text:
                 leftshift = totwidth - linewidth
             else:  # halign = 'left'
                 leftshift = 0
-            for glyph, x in lineglyphs:
-                word.append(glyph.place(x+leftshift, yvals[lineidx], self.size))
+            for glyph, x, dy in lineglyphs:
+                elm = glyph.place(x+leftshift, yvals[lineidx]+dy, self.size)
+                if elm is not None:
+                    word.append(elm)
 
-        ymin = yvals[0] - self.font.info.layout.ascent*scale
-        ymax = yvals[-1] - self.font.info.layout.descent*scale
+        ymin = yvals[0] - self.font.info.layout.ymax*scale
+        ymax = yvals[-1] - self.font.info.layout.ymin*scale
 
         if not config.svg2:
             symbols = []
-        return Symbols(word, symbols, totwidth, ymin, ymax)
+        return Symbols(word, symbols, totwidth, xmin, ymin, ymax)
 
     def getsize(self) -> tuple[float, float]:
         ''' Calculate width and height (including ascent/descent) of string '''
@@ -567,51 +635,3 @@ class Text:
     def getyofst(self) -> float:
         ''' Y-shift from bottom of bbox to 0 '''
         return -self._symbols.ymax
-
-
-def _build_fontlist():
-    ''' Generate list of system fonts locations and their names '''
-    if sys.platform.startswith('win'):
-        paths = [Path(r'C:\Windows\Fonts'),
-                 Path(os.path.expandvars(r'%APPDATA%\Microsoft\Windows\Fonts'))]
-    elif sys.platform.startswith('darwin'):
-        paths = [Path().home() / 'Library/Fonts',
-                 Path('/Library/Fonts/'),
-                 Path('/System/Library/Fonts')]
-    else:
-        paths = [Path('/usr/share/fonts/'),
-                 Path('/usr/local/share/fonts/'),
-                 Path().home() / '.fonts',
-                 Path().home() / '.local/share/fonts']
-
-    fontlist = []
-    for p in paths:
-        fontlist.extend(p.rglob('*.ttf'))
-        fontlist.extend(p.rglob('*.otf'))
-
-    findex = {}
-    for fname in fontlist:
-        try:
-            f = Font(fname)
-        except ValueError:
-            continue  # Unsupported Font
-        family = f.info.names.family.lower()
-        subfamily = f.info.names.subfamily.lower()
-        subfamily = subfamily.replace('book', 'regular').replace('normal', 'regular')
-        subfamily = subfamily.replace('italique', 'italic').replace('gras', 'bold')
-        findex[(family, subfamily)] = fname
-    return findex
-
-
-def findfont(name, style='Regular'):
-    ''' Find a font file by name '''
-    if name is None:
-        return None
-
-    if Path(name).exists():
-        return Path(name)
-
-    global fontindex
-    if fontindex is None:
-        fontindex = _build_fontlist()
-    return fontindex.get((name.lower(), style.lower()), None)
